%    File: treeprf.pl
%  Author: E.P. Stabler
% Updated: March 1995
% Purpose: just like showprf, except show_step/4 asserts each step
%             and derivation_tree/1 is defined to collect tree from them

:- module(treeprf,[
	parse_proof/3,
	glc_proof/6,
	show_step/4,
	derivation_tree/2,
	show_glc_proof/1,
	show_seq/1,
	show_abbr_seq/1,
	abbreviate_rule/2,
	abbreviate_seq/2,
	display_tuples/2]).

:- op(710,xfy,:->).	% for rewrite rules
:- op(720,xfy,:=>).	% for rewrite rules with prefix function
:- op(710,fx,:->).	% for predictions in states
:- op(910,fx,~).	% for predicted categories
:- op(800,xfy,@).	% for substitutions

:- use_module(step, [step/0]).			% for trace

:- dynamic derivation_step/4.

parse_proof(Input, Tree, Proof) :- 
	parse_proof(Input, Item, Cat, Proof),
	user:cat_from_item(Item, Cat),	% parse-method-specific
	user:tree_from_cat(Cat, Tree).	% grammar-specific

parse_proof(Input, Item, Cat, ProofTree) :-
	glc_proof([]*Input, [Item]*[], Cat, 1, []:[ax], ProofTree).

show_step(StepNo, Name:(A:->RHS), Sequent1, abbreviate) :- !,
	abbreviate_rule(A:->RHS, Rule), 
	write(StepNo), write('. '), write(Name), write(': '), print(Rule), nl,
	show_abbr_seq(Sequent1),
	retractall(derivation_step(StepNo,_Name,_Rule,_Sequent1)),
	assert(derivation_step(StepNo,Name,Rule,Sequent1)),
	step.
show_step(StepNo, Name:(A:=>RHS), Sequent1, abbreviate) :- !,
	abbreviate_rule(A:=>RHS, Rule), 
	write(StepNo), write('. '), write(Name), write(': '), print(Rule), nl,
	show_abbr_seq(Sequent1),
	retractall(derivation_step(StepNo,_Name,_Rule,_Sequent1)),
	assert(derivation_step(StepNo,Name,Rule,Sequent1)),
	step.
show_step(StepNo, Rule, Sequent1, abbreviate) :- !,
	write(StepNo), write('. '), print(Rule), nl,
	show_abbr_seq(Sequent1),
	retractall(derivation_step(StepNo,_Name,_Rule,_Sequent1)),
	assert(derivation_step(StepNo,none,Rule,Sequent1)),
	step.
show_step(StepNo, Rule, Sequent1, _) :-
	write(StepNo), write('. '), print(Rule), nl,
	show_seq(Sequent1),
	retractall(derivation_step(StepNo,_Name,_Rule,_Sequent1)),
	assert(derivation_step(StepNo,none,Rule,Sequent1)),
	step.

glc_proof(Sequent, Sequent, _, _, Trees:Rule, (Rule:Sequent)/Trees).
glc_proof(Sequent0, Sequent, Cat, Step, Trees:Rule0, ProofTree) :-
	user: infer(Sequent0, Sequent1, Rule, Cat),
	show_step(Step, Rule, Sequent1, full),
%	show_step(Step, Rule, Sequent1, abbreviate),
	Step1 is Step+1,
	premises((Rule0:Sequent0), Trees, Premises),
	glc_proof(Sequent1, Sequent, Cat, Step1, Premises:Rule, ProofTree).

premises(((Name:Side):Sequent), Trees, ['$m'('     ')/[],(Name:Sequent)/Trees, (Side/[]) ]):-!.
premises((Name:Sequent), Trees, [(Name:Sequent)/Trees]).

show_glc_proof([]).
show_glc_proof([Rule:Seq|Seqs]) :-
	portray_clause(Rule),
	show_seq(Seq),
	show_glc_proof(Seqs).

show_seq([] * Cons) :-
	write(emptystack), nl,
	write(' * '), write(Cons), nl.
show_seq([State|States] * Cons) :-
	show_states([State|States],full),
	write(' * '), write(Cons), nl.

show_abbr_seq([] * Cons) :-
	write([]), nl,
	write(' * '), write(Cons), nl.
show_abbr_seq([State|States] * Cons) :-
	show_states([State|States],abbr),
	write(' * '), write(Cons), nl.

show_states([], _).
show_states([State|States], Flag) :-
	show_state(State, Flag),
	show_states(States, Flag).

show_state([], _) :- !,
	write({}), nl.
show_state([Item|Items], Flag) :- !,
	write('{'),
	show_items([Item|Items], Flag).
show_state(A, full) :- 
	put(32), print(A), nl.
show_state(A, abbr) :- 
	put(32), abbreviate_rule(A, AbbrA), print(AbbrA), nl.

show_items([Item@Value], _Flag) :- !, 
	put(32), print(Item), write('@'),
	show_subst(Value), nl,
	put(32), write('}'), nl.
show_items([Item@Value|Items], Flag) :- !,
	put(32), print(Item), write('@'),
	show_subst(Value), write(','), nl,
	show_items(Items, Flag).

show_items([Item], full) :- !, 
	put(32), print(Item),
	put(32), write('}'), nl.
show_items([Item|Items], full) :- !,
	put(32), print(Item), write(','), nl,
	show_items(Items, full).

show_items([Item], abbr) :- !, 
	put(32), abbreviate_rule(Item, AbbrItem), print(AbbrItem),
	put(32), write('}'), nl.
show_items([Item|Items], abbr) :- 
	put(32), abbreviate_rule(Item, AbbrItem), print(AbbrItem),
	write(','), nl,
	show_items(Items, abbr).

show_subst((A,B)) :- !,
	show_subst(A), write(','),
	show_subst(B).
show_subst(A) :- !,
	nl, put(32), put(32), print(A).

% rule abbreviation
abbreviate_rule(( :-> LHS0-RHS0), ( :-> LHS-RHS)) :- !,
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).
abbreviate_rule((A0 :-> LHS0-RHS0), (A :-> LHS-RHS)) :- !,
	abbreviate(A0, A),
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).
abbreviate_rule((A0 :=> LHS0>RHS0), (A :=> LHS>RHS)) :- !,
	abbreviate(A0, A),
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).
abbreviate_rule((A0 :=> LHS0<RHS0), (A :=> LHS<RHS)) :- !,
	abbreviate(A0, A),
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).
abbreviate_rule([ :-> LHS0-RHS0], [ :-> LHS-RHS]) :- !, %"covered" predict
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).
abbreviate_rule(~A0, ~A) :- !,
	abbreviate_rule(A0, A).
abbreviate_rule(A/[], A) :- !.
abbreviate_rule(A/[_|_], A) :- !.
abbreviate_rule(A0, A) :-
	abbreviate(A0, A).

abbreviate_seq((LHS0*RHS0), (LHS*RHS)) :-
	abbreviates(LHS0, LHS),
	abbreviates(RHS0, RHS).

abbreviates([], []).
abbreviates([~Cat0|Cats], [~Cat|Abbrevs]) :- !,
	abbreviate(Cat0, Cat),
	abbreviates(Cats, Abbrevs).
abbreviates([Cat0|Cats], [Cat|Abbrevs]) :-
	abbreviate(Cat0, Cat),
	abbreviates(Cats, Abbrevs).

abbreviate(Cat, Functor) :- nonvar(Cat), functor(Cat, Functor, _Arity).

list2and([H],H) :- !.
list2and([H1,H2|T],(H1,And)) :- list2and([H2|T],And).
list2and([],true).
