/* tuple_tree.pl
 * 
 * transform a derivation tree to tuple notation and then,
 * elaborating qtree.pl, format the tree for the latex qtree.sty package
 * 
 * E. Stabler, Aug 2002
 */

:- module(tuple_tree,[tupleTree/2]).

tupleTree(Term,Licensors) :-
	tell('qtree.tex'),
	numbervars(Term, 0, _),
	write('\\documentclass{article}'),nl,
	write('\\usepackage{amsmath}'),nl,
	write('\\usepackage{qtree}'),nl,
	write('\\textwidth7in'),nl,
	write('\\textheight8.4in'),nl,
	write('\\oddsidemargin-.25in'),nl,
	write('\\evensidemargin-.25in'),nl,
	write('\\topmargin-.25in'),nl,
	write('\\headheight20pt'),nl,
	write('\\headsep30pt'),nl,
	write('\\footskip20pt'),nl,
	write('\\unitlength0.5mm'),nl,
	write('\\begin{document}'),nl,
	write('{\\tiny \\Tree '),nl,
	tupleTree(Term, 0, Licensors),
	nl,
	write('} \\end{document}'),nl,
	fail
	;
	told.
%	true.

tupleTree(X/[], Column, L) :- !,
	tab(Column), 
	write('{'),writeCat(X,L),!,write('}').
tupleTree(Cat/[Child|Ren], Column, L) :- !,
	tab(Column),write('[ '),
	tupleTrees([Child|Ren], Column, L),
	nl,write(' ].'),write('{'),writeCat(Cat,L),!,write('}').
tupleTree(X, Column, L) :- 
	tab(Column), 
	write('{'),writeCat(X,L),!,write('}').

tupleTrees([], _, _).
tupleTrees([Tree|Trees], Column, L) :-
	NextColumn is Column+4,
	nl, tupleTree(Tree, NextColumn, L),
	tupleRest_trees(Trees, NextColumn, L).

tupleRest_trees([], _, _).
tupleRest_trees([Tree|Trees], Column, L) :-
	nl, tupleTree(Tree, Column, L),
	tupleRest_trees(Trees, Column, L).

writeCat([(String::Fs0)|Chains],L) :-
	length(L,N0),
	N is N0+1,
	length(Strings,N),
	length(Categories,N),
	nth0(0,Strings,String),
	renameLexFs(Fs0,L,Fs),
	nth0(0,Categories,Fs),
	fillChains(Chains,Strings,Categories,L),
	writeMatrix(Strings,Categories).
writeCat([(String:Fs0)|Chains],L) :-
	length(L,N0),
	N is N0+1,
	length(Strings,N),
	length(Categories,N),
	nth0(0,Strings,String),
	renameDerFs(Fs0,L,Fs),
	nth0(0,Categories,Fs),
	fillChains(Chains,Strings,Categories,L),
	writeMatrix(Strings,Categories).
writeCat(Cat,_) :-
	write(Cat).

writeMatrix(Strings,Categories) :-
	write('$\\begin{pmatrix}'),nl,
	writeMatrixRow(Categories),
	writeMatrixRow(Strings),
	write('\\end{pmatrix}$'),nl.

writeMatrixRow([E,F|L]) :- !,
	(   var(E) -> 	write('\\epsilon & ')
	;   writeEntry(E), write(' & ')
	),
	writeMatrixRow([F|L]).
writeMatrixRow([E]) :- !,
	(   var(E) -> 	write('\\epsilon\\\\'), nl
	;   writeEntry(E), write('\\\\'), nl
	).
writeMatrixRow([]) :- write('\\\\'), nl.

fillChains([],_Strings,_Categories,_L).
fillChains([(String:[-F0|Fs0])|Chains],Strings,Categories,L) :-
	nth1(N,L,F0),
	renameDerFs([-F0|Fs0],L,Fs),
	nth0(N,Categories,Fs),
	nth0(N,Strings,String),
	fillChains(Chains,Strings,Categories,L).

renameLexFs([=F0|Fs0],L,[F|Fs]) :- !,
	name(F0,ListF0),
	name(F,[47|ListF0]), /* right selector / */
	renameDerFs(Fs0,L,Fs).
renameLexFs(Fs0,L,Fs) :- renameDerFs(Fs0,L,Fs).

renameDerFs([],_,[]).
renameDerFs([=F0|Fs0],L,[F|Fs]) :- !,	
	name(F0,ListF0),
	append(ListF0,"$\\backslash$",List),
	name(F,List), /* left selector \, doubled for correct printing */
	renameDerFs(Fs0,L,Fs).
renameDerFs([+F0|Fs0],L,[+N|Fs]) :- !,
	nth1(N,L,F0),
	renameDerFs(Fs0,L,Fs).
renameDerFs([-F0|Fs0],L,[-N|Fs]) :- !,
	nth1(N,L,F0),
	renameDerFs(Fs0,L,Fs).
renameDerFs([F0|Fs0],L,[F0|Fs]) :-
	renameDerFs(Fs0,L,Fs).

writeEntry([]) :- write('{\\epsilon}'), nl.
writeEntry([H|T]) :-
	write('\\hbox{'),
	writeEntryElements([H|T]),
	write('}').

writeEntryElements([E,F|T]) :- !, writeEntryElement(E), write('\\ '), writeEntryElements([F|T]).
writeEntryElements([E]) :- !, writeEntryElement(E).
writeEntryElements([]).

writeEntryElement(+E) :- !, write('{\\tt+'),write(E),write('}').
writeEntryElement(-E) :- !, write('{\\tt-'),write(E),write('}').
writeEntryElement(E) :- write(E).

